"""
Base module for flask_trace_util
"""
from .flask_trace import (
    _FlaskTrace,
    gcloud_trace_delegator,
    gcloud_trace_extractor,
    simple_user_id_extractor,
    simple_user_id_delegator,
)

gcloud_trace = _FlaskTrace(
    delegator=gcloud_trace_delegator,
    user_id_extractor=simple_user_id_extractor,
    user_id_delegator=simple_user_id_delegator,
)



"""
Main class for trace operations
"""
from flask import g, request
import requests
from flask_trace_util.util import extract_gcloud_trace


class _FlaskTrace:
    def __init__(self, **kwargs):
        self.app = None
        self.extractor = kwargs.get("extractor")
        self.delegator = kwargs.get("delegator")
        self.user_id_extractor = kwargs.get("user_id_extractor")
        self.user_id_delegator = kwargs.get("user_id_delegator")

    def init_app(self, app, **kwargs):
        """
            Init app
        """
        self.app = app
        self.extractor = kwargs.get("extractor") or self.extractor
        self.delegator = kwargs.get("delegator") or self.delegator
        self.user_id_extractor = (
            kwargs.get("user_id_extractor") or self.user_id_extractor
        )
        self.user_id_delegator = (
            kwargs.get("user_id_delegator") or self.user_id_delegator
        )

        if self.app is None:
            raise Exception("app cant be None")

        if self.extractor is None:
            raise Exception("extractor cant be None")

        if self.delegator is None:
            raise Exception("delegator cant be None")

        @app.before_request
        def init_session():  # pylint: disable=unused-variable
            g.trace = self.extractor()
            if self.user_id_extractor:
                g.user_id = self.user_id_extractor()

        @app.after_request
        def destroy_session(
            response,
        ):  # pylint: disable=unused-variable,unused-argument
            g.trace = None
            return response

    def request(self, method, url, **kwargs):
        """Proxy request function which uses extracted trace field

        Arguments:
            method {string} -- http method
            url {string} -- url

        Returns:
            object -- requests response object
        """
        if "headers" not in kwargs:
            kwargs["headers"] = {}
        header_name, header_value = self.delegator()
        kwargs["headers"][header_name] = header_value
        if self.user_id_delegator:
            u_header_name, u_header_value = self.user_id_delegator()
            kwargs["headers"][u_header_name] = u_header_value
        return requests.request(method, url, **kwargs)


def gcloud_trace_extractor(project_id):
    """Trace extractor for google cloud GKE

    Arguments:
        project_id {str} -- project id
    """

    def extractor():
        trace = request.headers.get("X-Cloud-Trace-Context")
        _, trace = extract_gcloud_trace(trace, project_id)
        return trace

    return extractor


def gcloud_trace_delegator():
    """Trace delegator for google cloud GKE

    Arguments:
        trace {str} -- trace
    """

    def delegator():
        trace = ""
        if g.trace:
            trace = g.trace.split("/")[-1]
        return "X-Cloud-Trace-Context", trace

    return delegator


def simple_user_id_extractor():
    """Simple user_id extractor
    """

    def extractor():
        user_id = request.headers.get("X-User-Id")
        if not user_id:
            user_id = "Anonymous"
        return user_id

    return extractor


def simple_user_id_delegator():
    """Simple user_id delegator
    """

    def delegator():
        user_id = "Anonymous"
        if g.user_id:
            user_id = g.user_id
        return "X-User-Id", user_id

    return delegator



"""
Common utilities for internal usage
"""


def extract_gcloud_trace(header, project_id):
    """
    Extract trace header from X-Cloud-Trace-Context
    """
    trace = ""
    if not header:
        trace = ""
    else:
        trace = f"projects/{project_id}/traces/{header.split('/')[0]}"
    return ("logging.googleapis.com/trace", trace)



"""
Contains Gunicorn specific loggers
"""
import json
from datetime import datetime
import traceback
from gunicorn.glogging import Logger
from flask_trace_util.util import extract_gcloud_trace


class GunicornLogger(Logger):
    """Extended gunicorn acces logger

    Arguments:
        Logger {object} -- [GunicornLogger]
    """

    exclude_routes = ["/"]
    include_request_headers = True
    include_response_headers = True
    _log_format_pairs = {
        "remote_address": "%(h)s",
        "timestamp": "%(t)s",
        "message": "%(r)s",
        "path": "%(U)s",
        "query": "%(q)s",
        "status": "%(s)s",
        "response_size": "%(B)s",
        "referer": "%(f)s",
        "user_agent": "%(a)s",
        "request_time": "%(L)s",
        "method": "%(m)s",
    }

    def access(self, resp, req, environ, request_time):
        """
        Access log function
        """
        if req.path in self.exclude_routes:
            return
        log_format = self._log_format_pairs.copy()
        if self.include_request_headers:
            log_format["request_headers"] = {}
            for k, _ in req.headers:
                log_format["request_headers"][k.lower()] = f"%({{{k.lower()}}}i)s"

        if self.include_response_headers:
            log_format["response_headers"] = {}
            for k, _ in resp.headers:
                log_format["response_headers"][k.lower()] = f"%({{{k.lower()}}}o)s"

        if hasattr(self, "get_custom_variables"):
            cvars = self.get_custom_variables(resp, req)
            for k, value in cvars.items():
                log_format[k.lower()] = f"%({{{k.lower()}}}e)s"
                environ[k.lower()] = value
        access_log_format = json.dumps(log_format)

        if not (
            self.cfg.accesslog
            or self.cfg.logconfig
            or self.cfg.logconfig_dict
            or (self.cfg.syslog and not self.cfg.disable_redirect_access_to_syslog)
        ):
            return
        atoms = self.atoms(resp, req, environ, request_time)
        atoms["t"] = str(datetime.utcnow())
        safe_atoms = self.atoms_wrapper_class(atoms)

        try:
            self.access_log.info(access_log_format, safe_atoms)
        except:
            self.error(traceback.format_exc())


class GCloudGunicornAccesLogger(GunicornLogger):
    """ Access logger for flask, gunicorn and google cloud

    Arguments:
        GunicornLogger {type} -- flask trace util GunicornLogger
    """

    project_id = None

    def get_custom_variables(self, resp, req):
        """Returns custom variables for acces logging

        Returns:
            dict -- custom env variables
        """
        cvars = {}
        for k, value in req.headers:
            if k.lower() == "X-Cloud-Trace-Context".lower():
                trace_key, trace_value = extract_gcloud_trace(value, self.project_id)
                cvars[trace_key] = trace_value
        return cvars
